#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/xmmsctrl.h>
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <wchar.h>
#include "tags.h"
#include "listento.h"
#include "gtkstuff.h"
#include "config.h"
#include "fmt.h"

static void init(void);
static void cleanup(void);
static void *xs_thread(void *);
static int going;

static pthread_t pt_listento;
static pthread_mutex_t m_listento = PTHREAD_MUTEX_INITIALIZER;

static GeneralPlugin xmms_listento =
{
	NULL,
	NULL,
	-1,
	NULL,
	init,
	about_show,
	configure_dialog,
	cleanup
};

static void init(void)
{
	char *username = NULL, *password = NULL;
	ConfigFile *cfgfile;
	going = 1;

	if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {
		xmms_cfg_read_string(cfgfile, "listen-to", "username",
				&username);
		xmms_cfg_read_string(cfgfile, "listen-to", "password",
				&password);
		xmms_cfg_free(cfgfile);
	}
	if ((!username || !password) || (!*username || !*password)) {
		pdebug("username/password not found - not starting");
		going = 0;
		return;
	}
	lt_init(username, password);
	free(username);
	free(password);
	if (pthread_create(&pt_listento, NULL, xs_thread, NULL)) {
		pdebug("Error creating thread");
		going = 0;
		return;
	}
	pdebug("plugin started");
}

static void cleanup(void)
{
	void *dummy;

	if (!going)
		return;
	pdebug("about to lock mutex");
	pthread_mutex_lock(&m_listento);
	pdebug("locked mutex");
	going = 0;
	pthread_mutex_unlock(&m_listento);
	pdebug("joining thread");
	pthread_join(pt_listento, &dummy);
}

static char ishttp(const char *a)
{
	char *tmp, *bp;
	int status = 0;

	if (!a || !*a)
		return 0;

	tmp = strdup(a);
	for (bp = tmp; *bp; bp++)
		*bp = toupper(*bp);
	if (strstr(tmp, "HTTP://"))
		status = -1;
	free(tmp);
	return status;
}

static void *xs_thread(void *data)
{
	int playing, pos, len, prevpos = -1, prevlen = -1,
		otime = 0, dosubmit, run = 1;

	while (run) {
		playing = xmms_remote_is_playing(0);
		if (!playing) {
			prevpos = -1;
			prevlen = -1;
		}
		pos = xmms_remote_get_playlist_pos(0);
		len = xmms_remote_get_playlist_time(0, pos);
		otime = xmms_remote_get_output_time(0);
		dosubmit = (otime/1000 > 120) ||
			(((double)otime/((double)len + 1) * 100) >= 10);
		if((pos != prevpos || len != prevlen) && playing && dosubmit) {
			char *fname, /**title, *artist,*/ *tmp = NULL, *sep, *tags;
			metadata meta;

			clean_meta(&meta);

			fname = xmms_remote_get_playlist_file(0, pos);
			prevpos = pos;
			prevlen = len;
			if (ishttp(fname)) {
				g_free(fname);
				continue;
			}
			tags = tag_exists(fname);
			if(tags[0] == CD_AUDIO)
			{
				ConfigFile *cfgfile;

				if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {
					tmp = strrchr(fname, '.');
					*tmp = '\0';
					tags[1] = (char)atoi(tmp - 2);
					pdebug(fmt_vastr("Track: %d", tags[1]));
					xmms_cfg_read_string(cfgfile, "CDDA", "device", &fname);
					xmms_cfg_free(cfgfile);
					pdebug(fmt_vastr("CD Device: %s", fname));
				}
			}
			if((tags[0] != NO_TAGS && (tags[0] & READ_ERROR) != READ_ERROR) || tags[1] != NO_TAGS)
			{
				meta = get_tag_data(fname, tags);
			}
			if(meta.artist != NULL && meta.title != NULL && meta.album != NULL)
			{
				meta.length = len / 1000;
				pdebug(fmt_vastr("submitting song"));
				lt_submitentry(meta);
			}
			else
				pdebug("couldn't determine artist - "
						"title - album, not submitting");
			g_free(tmp);
			g_free(fname);
			clean_meta(&meta);
		}
		pthread_mutex_lock(&m_listento);
		run = going;
		pthread_mutex_unlock(&m_listento);
		if (lt_idle()) {
			pdebug("Giving up due to fatal error");
			run = 0;
		}
		usleep(100000);
	}
	pdebug("thread: exiting");
	pthread_exit(NULL);
}

GeneralPlugin *get_gplugin_info(void)
{
	xmms_listento.description = g_strdup_printf("xmms listen-to %s", VERSION);
	return &xmms_listento;
}
